#
# SYNOPSIS
#
# AX_PRINTF_SIZE_T_FORMAT
#
# DESCRIPTION
#
#   This macro will define macros reminiscent of inttypes.h, with a
#   format PRI<placeholder>SIZE. In doing this, it first tries to
#   guess whether or not the host platform's printf() supports the z
#   modifier for one of its integer output placeholders. It will use
#   that if available. Otherwise, it will guess a compatible
#   modifier. It will also create the associated SCN<placeholder>SIZE
#   constants for scanf().
#
#   The following constants will set with AC_DEFINE():
#     - PRIdSIZE
#     - PRIiSIZE
#     - PRIoSIZE
#     - PRIuSIZE
#     - PRIxSIZE
#     - PRIXSIZE
#     - SCNdSIZE
#     - SCNiSIZE
#     - SCNoSIZE
#     - SCNuSIZE
#     - SCNxSIZE
#
#   The following cache variables will be set or used:
#    - ax_cv_printf_size_t_modifier: The conversion modifier to be
#        added before the `d' in `%d' to make printf() accept a -
#        size_t argument for that placeholder.
#
#   Example usage of one of these constants:
#       printf("There are %" PRIuSIZE " items in the list.\n", list_length);
#       printf("Item %04" PRIuSIZE " is interesting.\n", item_index);
#
# LICENSE
#
#   Copyright (C) 2011  Nathan Phillip Brink <ohnobinki@ohnopublishing.net>
#
#   Copying and distribution of this file, with or without modification,
#   are permitted in any medium without royalty provided the copyright
#   notice and this notice are preserved.  This file is offered as-is,
#   without any warranty.

#serial 1

AC_DEFUN([AX_PRINTF_SIZE_T_FORMAT],
[
	AC_REQUIRE([AC_CANONICAL_HOST])
	AC_REQUIRE([AC_PROG_CC_C99])

	AS_VAR_SET_IF([ax_cv_printf_size_t_modifier],
		[ax_printf_size_t_modifier_set=yes],
		[ax_printf_size_t_modifier_set=no])
	AS_IF([test "x$ax_printf_size_t_modifier_set" = "xyes"],
		[:],
		[test "x$host_os" = "xmingw32" -o "x$ac_cv_prog_cc_c99" = "xno"],
		[
		AC_CHECK_SIZEOF([size_t])
		AS_IF([test "x$ac_cv_sizeof_size_t" != "x0"],
			[
			# Try to guess the size relative to
			# printf()-friendly types since we were able
			# to calculate the size of size_t.
			modifier_found=no

			AC_CHECK_SIZEOF([short])
			AS_IF([test "x$ac_cv_sizeof_size_t" = "x$ac_cv_sizeof_short"],
				[
				ax_printf_size_t_modifier=h
				modifier_found=yes
				],
				[AC_CHECK_SIZEOF([int])])
			AS_IF([test "x$ac_cv_sizeof_size_t" = "x$ac_cv_sizeof_int"],
				[
				ax_printf_size_t_modifier=
				modifier_found=yes
				],
				[
				AS_IF([test "x$modifier_found" = "xno"],
					[AC_CHECK_SIZEOF([long])])
				])
			AS_IF([test "x$ac_cv_sizeof_size_t" = "x$ac_cv_sizeof_long"],
				[
				ax_printf_size_t_modifier=l
				modifier_found=yes
				],
				[
				AS_IF([test "x$modifier_found" = "xno"],
					[AC_CHECK_SIZEOF([long long])])
				])
			AS_IF([test "x$ac_cv_sizeof_size_t" = "x$ac_cv_sizeof_long_long"],
				[
				ax_printf_size_t_modifier=ll
				modifier_found=yes
				])
			],
			[
			# Insane fallback if we don't know size_t's width.
			ax_printf_size_t_modifier=l
			])
		],
		[ax_printf_size_t_modifier=z])

	AC_CACHE_CHECK([printf() modifier to use for (s)size_t],
		[ax_cv_printf_size_t_modifier],
		[ax_cv_printf_size_t_modifier="$ax_printf_size_t_modifier"])

	ax_printf_d_size_format="$ax_cv_printf_size_t_modifier"d
	AC_DEFINE_UNQUOTED([PRIdSIZE], ["$ax_printf_d_size_format"], [printf() %d modifier for ssize_t])
	AC_DEFINE_UNQUOTED([SCNdSIZE], ["$ax_printf_d_size_format"], [scanf() %d modifier for ssize_t])

	ax_printf_i_size_format="$ax_cv_printf_size_t_modifier"i
	AC_DEFINE_UNQUOTED([PRIiSIZE], ["$ax_printf_i_size_format"], [printf() %i modifier for ssize_t])
	AC_DEFINE_UNQUOTED([SCNiSIZE], ["$ax_printf_i_size_format"], [scanf() %i modifier for ssize_t])

	ax_printf_o_size_format="$ax_cv_printf_size_t_modifier"o
	AC_DEFINE_UNQUOTED([PRIoSIZE], ["$ax_printf_o_size_format"], [printf() %o modifier for size_t])
	AC_DEFINE_UNQUOTED([SCNoSIZE], ["$ax_printf_o_size_format"], [scanf() %o modifier for size_t])

	ax_printf_u_size_format="$ax_cv_printf_size_t_modifier"u
	AC_DEFINE_UNQUOTED([PRIuSIZE], ["$ax_printf_u_size_format"], [printf() %u modifier for size_t])
	AC_DEFINE_UNQUOTED([SCNuSIZE], ["$ax_printf_u_size_format"], [scanf() %u modifier for size_t])

	ax_printf_x_size_format="$ax_cv_printf_size_t_modifier"x
	AC_DEFINE_UNQUOTED([PRIxSIZE], ["$ax_printf_x_size_format"], [printf() %x modifier for size_t])
	AC_DEFINE_UNQUOTED([SCNxSIZE], ["$ax_printf_x_size_format"], [scanf() %x modifier for size_t])

	ax_printf_X_size_format="$ax_cv_printf_size_t_modifier"X
	AC_DEFINE_UNQUOTED([PRIXSIZE], ["$ax_printf_X_size_format"], [printf() %X modifier for size_t])
])
