/*
 * Copyright (C) 2011 Nathan Phillip Brink <ohnobinki@ohnopublishing.net>
 *
 * This file is a part of difffilter.
 *
 * difffilter is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * difffilter is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with difffilter.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "block.h"

#include <inttypes.h>
#include <list.h>
#include <stdio.h> /* snprintf() */
#include <stdlib.h> /* malloc() */
#include <string.h> /* strdup() */
#include <unistd.h> /* ssize_t */

/* locals */

struct diff_block
{
  list_t lines;

  unsigned long new_begin;
  unsigned long orig_begin;

  char *comment;
  size_t comment_len;
};

diff_block_t diff_block_init(size_t orig_begin, size_t new_begin)
{
  diff_block_t diff_block;

  diff_block = malloc(sizeof(struct diff_block));
  if(!diff_block)
    return NULL;

  diff_block->lines = list_init();
  if(!diff_block->lines)
    {
      free(diff_block);
      return NULL;
    }

  diff_block->orig_begin = orig_begin;
  diff_block->new_begin = new_begin;

  diff_block->comment = NULL;
  diff_block->comment_len = 0;

  return diff_block;
}

int diff_block_add(diff_block_t diff_block, enum diff_line_type line_type, const char *line_content)
{
  struct diff_line *diff_line;

  diff_line = malloc(sizeof(struct diff_line));
  if(!diff_line)
    return 1;

  diff_line->type = line_type;
  diff_line->line = strdup(line_content);
  diff_line->len = strlen(line_content);

  list_insert_after(diff_block->lines, diff_line, 0);

  return 0;
}

int diff_block_interpret(diff_block_t diff_block, const char *line)
{
  enum diff_line_type line_type;

  switch(line[0])
    {
    case ' ':
      line_type = DIFF_LINE_NOOP;
      break;

    case '+':
      line_type = DIFF_LINE_PLUS;
      break;

    case '-':
      line_type = DIFF_LINE_MINUS;
      break;

    default:
      /**
       * This will catch 0-length strings too, so dereferencing
       * line[1] below is legal.
       */
      return 1;
    }

  return diff_block_add(diff_block, line_type, &line[1]);
}

void diff_block_begin_get(diff_block_t diff_block, size_t *orig_begin, size_t *new_begin)
{
  if (orig_begin)
    *orig_begin = diff_block->orig_begin;
  if (new_begin)
    *new_begin = diff_block->new_begin;
}

void diff_block_begin_set(diff_block_t diff_block, size_t orig_begin, size_t new_begin)
{
  diff_block->orig_begin = orig_begin;
  diff_block->new_begin = new_begin;
}

int diff_block_comment_set(diff_block_t diff_block, const char *comment)
{
  free(diff_block->comment);

  diff_block->comment = NULL;
  diff_block->comment_len = 0;
  if (comment)
    diff_block->comment_len = strlen(comment);

  if (diff_block->comment_len)
    {
      diff_block->comment = strdup(comment);
      if (!diff_block->comment)
	{
	  diff_block->comment_len = 0;
	  return 1;
	}
    }

  return 0;
}

const char *diff_block_comment_get(diff_block_t diff_block)
{
  return diff_block->comment;
}

list_t diff_block_lines(const diff_block_t diff_block)
{
  return diff_block->lines;
}


struct diff_block_reconstruct
{
  size_t orig_len;
  size_t new_len;

  size_t orig_pos;
  size_t new_pos;

  char *orig;
  char *new;
};
int diff_block_reconstruct_size_traverse(struct diff_block_reconstruct *reconstruct,
					 struct diff_line *line)
{
  switch (line->type)
    {
    case DIFF_LINE_NOOP:
      reconstruct->orig_len += line->len;
      reconstruct->new_len += line->len;
      break;
    case DIFF_LINE_PLUS:
      reconstruct->new_len += line->len;
      break;
    case DIFF_LINE_MINUS:
      reconstruct->orig_len += line->len;
      break;
    }

  return TRUE;
}

int diff_block_reconstruct_cat(struct diff_block_reconstruct *reconstruct,
			       struct diff_line *line)
{
  switch (line->type)
    {
    case DIFF_LINE_NOOP:
    case DIFF_LINE_PLUS:
      memcpy(reconstruct->new + reconstruct->new_pos, line->line, line->len);
      reconstruct->new_pos += line->len;

      if (line->type == DIFF_LINE_PLUS)
	break;
    case DIFF_LINE_MINUS:
      memcpy(reconstruct->orig + reconstruct->orig_pos, line->line, line->len);
      reconstruct->orig_pos += line->len;
      break;
    }

  return TRUE;
}

int diff_block_reconstruct(diff_block_t diff_block,
			   char **orig,
			   size_t *orig_len,
			   char **new,
			   size_t *new_len)
{
  struct diff_block_reconstruct reconstruct;

  reconstruct.orig_len = 0;
  reconstruct.new_len = 0;
  list_traverse(diff_block->lines, &reconstruct,
		(list_traverse_func_t)&diff_block_reconstruct_size_traverse,
		LIST_FORW|LIST_SAVE|LIST_FRNT);

  *orig = malloc(reconstruct.orig_len + 1);
  *new = malloc(reconstruct.new_len + 1);
  if (!*orig
      || !*new)
    {
      free(*orig);
      free(*new);
      return 1;
    }

  reconstruct.orig = *orig;
  reconstruct.new = *new;
  reconstruct.orig_pos = 0;
  reconstruct.new_pos = 0;

  list_traverse(diff_block->lines, &reconstruct,
		(list_traverse_func_t)&diff_block_reconstruct_cat,
		LIST_FORW|LIST_SAVE|LIST_FRNT);
  if (reconstruct.orig_pos != reconstruct.orig_len
      || reconstruct.new_pos != reconstruct.new_len)
    {
      fprintf(stderr, "diff_block_reconstruct(): Off-by-one error: orig_pos = %" PRIuSIZE ", orig_len = %" PRIuSIZE "; new_pos = %" PRIuSIZE ", new_len = %" PRIuSIZE "\n",
	      reconstruct.orig_pos, reconstruct.orig_len,
	      reconstruct.new_pos, reconstruct.new_len);
      free(*orig);
      free(*new);
      return 1;
    }

  /*
   * Kindly, but unnecessarily, NULL-terminate the buffers.
   */
  (*orig)[reconstruct.orig_len] = '\0';
  (*new)[reconstruct.new_len] = '\0';

  *orig_len = reconstruct.orig_len;
  *new_len = reconstruct.new_len;

  return 0;
}

struct diff_block_render_size_traverse
{
  size_t len;
  size_t new_numlines;
  size_t orig_numlines;
};

int diff_block_render_size_traverse(struct diff_block_render_size_traverse *size_traverse, struct diff_line *diff_line)
{
  size_traverse->len += 1 + diff_line->len;
  switch(diff_line->type)
    {
    case DIFF_LINE_NOOP:
      size_traverse->new_numlines ++;
      size_traverse->orig_numlines ++;
      break;

    case DIFF_LINE_PLUS:
      size_traverse->new_numlines ++;
      break;

    case DIFF_LINE_MINUS:
      size_traverse->orig_numlines ++;
      break;
    }

  return TRUE;
}

struct diff_block_render_traverse
{
  /**
   * The position where the next byte should be written.
   */
  char *outbuf_ptr;
  /**
   * The position beyond which no more bytes should be written.
   */
  char *outbuf_end;
};

int diff_block_render_traverse(struct diff_block_render_traverse *render_traverse, struct diff_line *diff_line)
{
  char type_char;

  if(render_traverse->outbuf_ptr + 1 + diff_line->len > render_traverse->outbuf_end)
    return FALSE;

  switch(diff_line->type)
    {
    case DIFF_LINE_NOOP:
      type_char = ' ';
      break;

    case DIFF_LINE_PLUS:
      type_char = '+';
      break;

    case DIFF_LINE_MINUS:
      type_char = '-';
      break;

    default:
      /* explode. This should be unreachable. */
      type_char = '&';
      break;
    }

  *render_traverse->outbuf_ptr = type_char;
  render_traverse->outbuf_ptr ++;

  memcpy(render_traverse->outbuf_ptr, diff_line->line, diff_line->len);
  render_traverse->outbuf_ptr += diff_line->len;

  /**
   * The caller of list_traverse() depends on LIST_EXTENT being thrown
   * so that it knows we successfully visited each item of the list.
   */
  return TRUE;
}

ssize_t diff_block_render(diff_block_t diff_block, char **buf)
{
  struct diff_block_render_size_traverse size_traverse;  
  struct diff_block_render_traverse traverse;

  int tmp;

  size_t orig_begin_translated;
  size_t new_begin_translated;

  size_traverse.len = 0;
  size_traverse.new_numlines = 0;
  size_traverse.orig_numlines = 0;
  tmp = list_traverse(diff_block->lines, &size_traverse, (list_traverse_func_t)&diff_block_render_size_traverse, LIST_FRNT|LIST_FORW|LIST_SAVE);
  if(tmp != LIST_EXTENT)
    {
      return 1;
    }

  /*
   * If a diff only contains inserts or only contains deletions, the
   * half of the magic diff header with a ",0" in it will point to the
   * previous line instead of to the line that it would point to if a
   * line were being replaced. Thus, we have to sorta decrement in
   * these cases.
   */
  orig_begin_translated = diff_block->orig_begin;
  new_begin_translated = diff_block->new_begin;
  if (!size_traverse.orig_numlines)
    orig_begin_translated --;
  if (!size_traverse.new_numlines)
    new_begin_translated --;

  size_traverse.len += snprintf(NULL, 0, "@@ -%" PRIuSIZE,
				orig_begin_translated);
  if (size_traverse.orig_numlines != 1)
    size_traverse.len += snprintf(NULL, 0, ",%" PRIuSIZE,
				  size_traverse.orig_numlines);

  size_traverse.len += snprintf(NULL, 0, " +%" PRIuSIZE " @@\n",
				new_begin_translated);
  if (size_traverse.new_numlines != 1)
    size_traverse.len += snprintf(NULL, 0, ",%" PRIuSIZE,
				  size_traverse.new_numlines);

  if (diff_block->comment)
    size_traverse.len += 1 + diff_block->comment_len;

  /* add a null terminator to the end for convenience for now */
  *buf = malloc(size_traverse.len + 1);
  if(!*buf)
    return -1;

  traverse.outbuf_end = *buf + size_traverse.len;
  traverse.outbuf_ptr = *buf + snprintf(*buf, size_traverse.len, "@@ -%" PRIuSIZE,
					orig_begin_translated);
  if (size_traverse.orig_numlines != 1)
    traverse.outbuf_ptr += snprintf(traverse.outbuf_ptr, traverse.outbuf_end - traverse.outbuf_ptr, ",%" PRIuSIZE " ",
				    size_traverse.orig_numlines);
  else
    traverse.outbuf_ptr += snprintf(traverse.outbuf_ptr, traverse.outbuf_end - traverse.outbuf_ptr, " ");
  traverse.outbuf_ptr += snprintf(traverse.outbuf_ptr, traverse.outbuf_end - traverse.outbuf_ptr, "+%" PRIuSIZE,
				  new_begin_translated);
  if (size_traverse.new_numlines != 1)
    traverse.outbuf_ptr += snprintf(traverse.outbuf_ptr, traverse.outbuf_end - traverse.outbuf_ptr, ",%" PRIuSIZE,
				    size_traverse.new_numlines);
  traverse.outbuf_ptr += snprintf(traverse.outbuf_ptr, traverse.outbuf_end - traverse.outbuf_ptr, " @@");
  if (diff_block->comment)
    traverse.outbuf_ptr += snprintf(traverse.outbuf_ptr, traverse.outbuf_end - traverse.outbuf_ptr, " %s",
				    diff_block->comment);
  *traverse.outbuf_ptr = '\n';
  traverse.outbuf_ptr ++;

  tmp = list_traverse(diff_block->lines, &traverse, (list_traverse_func_t)&diff_block_render_traverse, LIST_FRNT|LIST_FORW|LIST_SAVE);
  if(tmp != LIST_EXTENT)
    {
      free(*buf);
      return -1;
    }

  (*buf)[size_traverse.len] = '\0';

  return size_traverse.len;
}

void diff_line_free(struct diff_line *diff_line)
{
  free(diff_line->line);
  free(diff_line);
}

void diff_block_free(diff_block_t diff_block)
{
  list_free(diff_block->lines, (list_dealloc_func_t)&diff_line_free);
  free(diff_block->comment);
  free(diff_block);
}
