/*
 * Copyright (C) 2011 Nathan Phillip Brink <ohnobinki@ohnopublishing.net>
 *
 * This file is a part of difffilter.
 *
 * difffilter is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * difffilter is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with difffilter.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef _DIFFFILTER_BLOCK_H
#define _DIFFFILTER_BLOCK_H

/**
 * \file
 * \brief Provides output buffering and an abstraction around elements
 *   of unified diffs.
 *
 * Currently, a diff line will _include_ the newline ('\n')
 * character. It may be decided in the future that this makes no sense
 * and amends will be made then.
 */

#include <list.h>
#include <unistd.h> /* ssize_t */

struct diff_block;
typedef struct diff_block *diff_block_t;

/**
 * \brief The different types of unified diff lines.
 */
enum diff_line_type
  {
    /**
     * \brief A no-op line, represented in unified diffs with a ' '.
     */
    DIFF_LINE_NOOP,
    /**
     * \brief An addition line, represented in unified diffs with a
     *   '+'.
     */
    DIFF_LINE_PLUS,
    /**
     * \brief A subtraction line, represented in unified diffs with a
     *   '-'.
     */
    DIFF_LINE_MINUS,
  };

struct diff_line
{
  enum diff_line_type type;
  /**
   *
   */
  char *line;
  /**
   * length of the line _excluding_ the '+', '-', or ' ' found at the
   * beginning of a diff line
   */
  size_t len;
};

/**
 * \brief Allocate memory for and initialize a block.
 *
 * \param orig_begin The first line of the original file that the
 *   patch touches.
 * \param new_begin The first line that the patch touches in the
 *   resultant file.
 */
diff_block_t diff_block_init(size_t orig_begin, size_t new_begin);

/**
 * \brief Add a line to a diff_block_t.
 *
 * \param diff_block The diff_block initialized with diff_block_init().
 * \param line_type The type of unified diff line.
 * \param line_content The content of the unified diff line. As
 *   line_type includes information normally give in the first
 *   character of the line, this character shouldn't be included in
 *   line.
 * \return 0 on success. 1 on failure.
 */
int diff_block_add(diff_block_t diff_block, enum diff_line_type line_type, const char *line_content);

/**
 * \brief
 *   Retrieve the beginning line numbers for the original and new
 *   portions of the diff block.
 *
 * \param diff_block The diff_block.
 * \param orig_begin Will be filled with the orig_begin. If NULL is given, then this is ignored.
 * \param new_begin Will be filled with the new_begin. If NULL is given, then this is ignored.
 */
void diff_block_begin_get(diff_block_t diff_block, size_t *orig_begin, size_t *new_begin);

/**
 * \brief
 *   Set the beginning line numbers for the original and new portions
 *   of the diff block.
 *
 * You must call this when deleting lines from or adding lines to the
 * beginning of the line list. You may add and remove lines from the
 * list you retrieved with diff_block_lines().
 *
 * \param diff_block The diff_block.
 * \param orig_begin The new orig_begin.
 * \param new_begin The new new_begin.
 */
void diff_block_begin_set(diff_block_t diff_block, size_t orig_begin, size_t new_begin);

/**
 * \brief
 *   Set a diff block comment.
 *
 * A diff block comment is found on the same line as a diff block's
 * header. Normally, it specifies what function the diff block is an
 * excerpt of in the case that the diff is against a file written in a
 * programing language which uses functions. For a python-patch
 * example:
 *
 * @@ -48,8 +48,7 @@ class MirrorStatus():
 *
 * When setting the block's comment field, do not include the spacing
 * before the comment. That will be factored in automatically.
 *
 * \param diff_block
 *   The diff block for which the comment is being set.
 * \param comment
 *   A C string with the comment to set or NULL to specify no
 *   comment. Setting a comment removes any previously-set comment,
 *   and NULL can be used to erase a previously-set comment. The
 *   string will be copied.
 * \return
 *   0 on success, 1 on error.
 */
int diff_block_comment_set(diff_block_t diff_block, const char *comment);

/**
 * \brief
 *   Retrieve a block's comment.
 *
 * \param diff_block
 *   The block whose comment should be retrieved.
 * \return
 *   NULL if there is no comment or a pointer to a C string which may
 *   not be modified and may not be accessed after any futher calls to
 *   diff_* functions involing this diff_block.
 */
const char *diff_block_comment_get(diff_block_t diff_block);

/**
 * \brief
 *   Gets the list which stores all of the diff lines.
 *
 * \return An editable list of diff lines. Each member is a struct
 *   diff_line. You may add, edit, and remove items from this list. Do
 *   not call list_free on this list. This list is invalid after
 *   calling diff_block_free(). You may not call any diff_block_
 *   functions from within a list traversal started by
 *   list_traverse().
 */
list_t diff_block_lines(const diff_block_t diff_block);

/**
 * \brief Interpret a unified diff line and add it to a diff_block_t.
 *
 * This function only handles individual addition, subtraction, or
 * no-op lines. It doesn't handle diff headers or magic diff headers,
 * as those constructs indicate that a new diff_block should be
 * started which is outside of the scope of diff_block itself.
 *
 * \param diff_block The diff_block initialized with diff_block_init().
 * \param line The line to interpret, including the first character
 *   which indicates the line type.
 * \return 0 on success. 1 on error, such as if the first character is
 *   invalid.
 */
int diff_block_interpret(diff_block_t diff_block, const char *line);

/**
 * \brief
 *   Reconstructs the portions of the original and new files
 *   represented by this block.
 *
 * This allows output similar to diff(1)'s *.rej files to be obtained.
 *
 * \param diff_block The diff_block.
 * \param orig Where to store the newly allocated buffer with the
 *   original file's contents. You must call free() on the value
 *   obtained here.
 * \param orig_len Will be filled with the number of bytes you may access from orig.
 * \param new Where to store the newly allocated buffer with the new
 *   file's contents. You must call free() on the value obtained here.
 * \param new_len Will be filled with the number of bytes you may access from new.
 * \return 0 on success. 1 on error. If 1 is returned, then don't
 *   trust or free() the values of *orig and *new.
 */
int diff_block_reconstruct(diff_block_t diff_block,
			   char **orig,
			   size_t *orig_len,
			   char **new,
			   size_t *new_len);

/**
 * \brief Render a diff_block_t.
 *
 * This function uses the convention of asprintf(), not getline(), for
 * its output buffer handling.
 *
 * \param diff_block The diff_block.
 * \param buf Where to store a newly allocated pointer to the results
 *   of rendering this block.
 * \return The length of buf or -1 on error.
 */
ssize_t diff_block_render(diff_block_t diff_block, char **buf);

/**
 * \brief
 *   Free a struct diff_line *diff_line.
 */
void diff_line_free(struct diff_line *diff_line);

/**
 * \brief Free a diff_block_t.
 */
void diff_block_free(diff_block_t diff_block);

#endif /* _DIFFFILTER_BLOCK_H */
