/*
 * Copyright (C) 2011 Nathan Phillip Brink <ohnobinki@ohnopublishing.net>
 *
 * This file is a part of difffilter.
 *
 * difffilter is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * difffilter is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with difffilter.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "block.h"
#include "diff.h"

#include <list.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

struct diff
{
  list_t diff_block_list;
  char *orig_filename;
  char *new_filename;
};

/**
 * \brief
 *   Move the current element of the diff_block_list to the vicinity
 *   of a line.
 *
 * If the sought for orig_begin is not found, the current element will
 * be positioned at an element whose orig_begin is prior to the given
 * orig_begin. If the list is empty, then nothing shall be done.
 *
 * If the list only has one element in it, then the list's current
 * element _may_ be set to an element where its orig_begin is greater
 * than the specified orig_begin.
 *
 * \param diff The diff.
 * \param orig_begin The line to seek for.
 */
static void diff_seek(diff_t diff, size_t orig_begin);

diff_t diff_init(const char *orig_filename, const char *new_filename)
{
  diff_t diff;

  diff = malloc(sizeof(struct diff));
  if(!diff)
    return NULL;

  diff->diff_block_list = list_init();
  diff->orig_filename = strdup(orig_filename);
  diff->new_filename = strdup(new_filename);
  if(!diff->diff_block_list
     || !diff->orig_filename
     || !diff->new_filename)
    {
      if(diff->diff_block_list)
	list_free(diff->diff_block_list, LIST_NODEALLOC);
      free(diff->orig_filename);
      free(diff->new_filename);
      free(diff);
      return NULL;
    }

  return diff;
}

int diff_add(diff_t diff, diff_block_t diff_block)
{
  diff_block_t curr_block;
  size_t curr_orig_begin;
  size_t orig_begin;

  if(!diff_block)
    return 1;

  if (list_empty(diff->diff_block_list))
    {
      list_insert_after(diff->diff_block_list, diff_block, 0);
      return 0;
    }

  diff_block_begin_get(diff_block, &orig_begin, NULL);
  diff_seek(diff, orig_begin);

  curr_block = list_curr(diff->diff_block_list);
  diff_block_begin_get(curr_block, &curr_orig_begin, NULL);
  if (orig_begin > curr_orig_begin)
    list_insert_after(diff->diff_block_list, diff_block, 0);
  else
    list_insert_before(diff->diff_block_list, diff_block, 0);

  return 0;
}

/**
 * \brief for diff_visit_blocks_traverse()'s state.
 */
struct diff_visit_blocks_traverse
{
  short delete_current_block;
  list_t diff_block_list;
  diff_visit_blocks_func_t func;
  void *func_data;
};

/**
 * \brief Traversal helper for diff_visit_blocks().
 */
int diff_visit_blocks_traverse(struct diff_visit_blocks_traverse *blocks_traverse, diff_block_t block)
{
  enum diff_visit_blocks_func_return tmp;

  tmp = (*blocks_traverse->func)(block, blocks_traverse->func_data);
  switch(tmp)
    {
    case DIFF_VISIT_BLOCKS_KEEP:
      return TRUE;

    case DIFF_VISIT_BLOCKS_DELETE:
      blocks_traverse->delete_current_block = 1;
      return FALSE;

    default:
      fprintf(stderr, "error: diff_visit_blocks_traverse(): Recieved invalid return value from callback function.");
      return TRUE;
    }
}

int diff_visit_blocks(diff_t diff, diff_visit_blocks_func_t func, void *data)
{
  struct diff_visit_blocks_traverse blocks_traverse;
  int list_state;

  blocks_traverse.diff_block_list = diff->diff_block_list;
  blocks_traverse.func = func;
  blocks_traverse.func_data = data;

  list_mvfront(diff->diff_block_list);

  list_state = LIST_OK;
  while(list_state != LIST_EXTENT
	/* it'd be nice if liblist didn't treat LIST_EMPTY as a special case */
#ifdef LIST_EMPTY
	&& list_state != LIST_EMPTY
#endif
	)
    {
      blocks_traverse.delete_current_block = 0;
      list_state = list_traverse(diff->diff_block_list,
				 &blocks_traverse,
				 (list_traverse_func_t)&diff_visit_blocks_traverse,
				 LIST_CURR|LIST_ALTR|LIST_FORW);

      if(list_state == LIST_OK
	 && blocks_traverse.delete_current_block)
	{
	  if(list_curr(diff->diff_block_list) == list_rear(diff->diff_block_list))
	    {
	      /*
	       * If we're removing the last item of the list, then
	       * list_remove_curr() will move the current element
	       * pointer to the previous list item and not the
	       * subsequent one. Thus, make sure our loop stops.
	       */
	      list_state = LIST_EXTENT;
	    }
	  diff_block_free(list_curr(diff->diff_block_list));
	  list_remove_curr(diff->diff_block_list);
	}
    }

  return 0;
}

/*
 * start of diff_render-related stuffage.
 */
struct diff_render_block
{
  char *buf;
  size_t len;
};

/**
 * \brief For use with list_free().
 */
void diff_render_traverse_blocks_free(struct diff_render_block *render_block)
{
  free(render_block->buf);
  free(render_block);
}

struct diff_render_traverse_blocks
{
  /* each list item if of type struct diff_render_block * */
  list_t render_block_list;
  size_t len;
};

/**
 * \brief List traverser for diff_render().
 */
int diff_render_traverse_blocks(struct diff_render_traverse_blocks *traverse_blocks, diff_block_t block)
{
  struct diff_render_block *render_block;
  ssize_t tmp;

  render_block = malloc(sizeof(struct diff_render_block));
  if(!render_block)
    return FALSE;

  tmp = diff_block_render(block, &render_block->buf);
  if(tmp == -1)
    {
      free(render_block);
      return FALSE;
    }

  /* overall length */
  traverse_blocks->len += tmp;
  /* per-block length */
  render_block->len = tmp;
  list_insert_after(traverse_blocks->render_block_list, render_block, 0);

  return TRUE;
}

struct diff_render_traverse_cat
{
  char *buf;
  size_t pos;
  size_t len;
};

/**
 * \brief List traverser for diff_render()'s second pass.
 */
int diff_render_traverse_cat(struct diff_render_traverse_cat *traverse_cat, struct diff_render_block *render_block)
{
  if(render_block->len + traverse_cat->pos > traverse_cat->len)
    return FALSE;

  memcpy(&traverse_cat->buf[traverse_cat->pos],
	 render_block->buf,
	 render_block->len);
  traverse_cat->pos += render_block->len;

  return TRUE;
}

ssize_t diff_render(diff_t diff, char **buf)
{
  struct diff_render_traverse_blocks traverse_blocks;
  struct diff_render_traverse_cat traverse_cat;

  int tmp;

  /*
   * If we have no sub blocks, don't render anything at all. But we're
   * still responsible to allocate a buffer.
   *
   * This is useful when diff_visit_blocks()-based filters end up
   * removing all subblocks.
   */
  if (list_empty(diff->diff_block_list))
    {
      /*
       * I've heard that malloc(0) can do weird things... but it's
       * still compatible with free() and since we return 0, nobody
       * should ever actually dereference this ;-)
       */
      *buf = malloc(0);
      return 0;
    }

  /*
   * Calculate the length of the final buffer and render individual
   * blocks.
   */

  traverse_blocks.render_block_list = list_init();
  traverse_blocks.len = 0;

  if(!traverse_blocks.render_block_list)
    return -1;

  tmp = list_traverse(diff->diff_block_list, &traverse_blocks, (list_traverse_func_t)&diff_render_traverse_blocks, LIST_FRNT|LIST_FORW|LIST_SAVE);
  /* if we don't get LIST_EXTENT, then the traversal was interrupted */
  if(tmp == LIST_OK)
    {
      /* Then the traversal was interrupted... */
      list_free(traverse_blocks.render_block_list, (list_dealloc_func_t)&diff_render_traverse_blocks_free);
      return -1;
    }

  /*
   * The length also includes our header.
   */
  traverse_blocks.len += strlen("--- ") + strlen(diff->orig_filename) + strlen("\n")
    + strlen("+++ ") + strlen(diff->new_filename) + strlen("\n");

  /*
   * Concatenate the individually rendered blocks into one big diff
   * buffer.
   */

  traverse_cat.pos = 0;
  traverse_cat.len = traverse_blocks.len;
  /*
   * we need a + 1 here because snprintf() always terminates strings
   * with '\0' even when unnecessary.
   */
  traverse_cat.buf = malloc(traverse_blocks.len + 1);
  if(!traverse_cat.buf)
    {
      list_free(traverse_blocks.render_block_list, (list_dealloc_func_t)&diff_render_traverse_blocks_free);
      return -1;
    }

  /*
   * The ``+ 1'' is because of how snprintf handles NULL bytes. Since
   * we're appending to the string that snprintf gives us, we sort of
   * ignore that ``+ 1'' elsewhere ;-). See the note for the malloc()
   * call above.
   */
  snprintf(&traverse_cat.buf[traverse_cat.pos],
	   strlen("--- \n") + strlen(diff->orig_filename) + 1,
	   "--- %s\n", diff->orig_filename);
  traverse_cat.pos += strlen("--- \n") + strlen(diff->orig_filename);

  snprintf(&traverse_cat.buf[traverse_cat.pos],
	   strlen("+++ \n") + strlen(diff->new_filename) + 1,
	   "+++ %s\n", diff->new_filename);
  traverse_cat.pos += strlen("+++ \n") + strlen(diff->new_filename);

  tmp = list_traverse(traverse_blocks.render_block_list, &traverse_cat, (list_traverse_func_t)&diff_render_traverse_cat, LIST_FRNT|LIST_FORW|LIST_SAVE);
  list_free(traverse_blocks.render_block_list, (list_dealloc_func_t)&diff_render_traverse_blocks_free);
  if(tmp == LIST_OK)
    {
      free(traverse_cat.buf);
      return -1;
    }

  if(traverse_cat.len != traverse_cat.pos)
    {
      fprintf(stderr, __FILE__ ":%d:diff_render(): Unexpected inconsistency, %lu != %lu\n",
	      __LINE__, (unsigned long)traverse_cat.len, (unsigned long)traverse_cat.pos);
      free(traverse_cat.buf);
      return -1;
    }

  *buf = traverse_cat.buf;

  return traverse_cat.len;
}

int diff_write_file(diff_t diff, FILE *out)
{
  /*
   * for now, just write out to a buffer and then flush that to
   * stdout. Later, we can avoid allocating some memory by more direct
   * output.
   */
  char *buf;
  ssize_t len;
  ssize_t tmp;
  int ret;

  ret = 0;

  len = diff_render(diff, &buf);
  if(len == -1)
    return 1;

  tmp = fwrite(buf, 1, len, out);
  if(tmp != len)
    {
      ret = 1;
    }

  free(buf);

  return ret;
}

void diff_free(diff_t diff)
{
  list_free(diff->diff_block_list, (list_dealloc_func_t)&diff_block_free);
  free(diff->orig_filename);
  free(diff->new_filename);
  free(diff);
}

struct diff_seek_traverse
{
  size_t orig_begin;
  int dir;
};

/**
 * \brief
 *   Traversal function for diff_seek().
 *
 * \param orig_begin A pointer to the sought line. If negative, this indicates 
 */
static int diff_seek_traverse(struct diff_seek_traverse *seek, diff_block_t block)
{
  size_t curr_orig_begin;

  diff_block_begin_get(block, &curr_orig_begin, NULL);
  if ((seek->dir == LIST_FORW && curr_orig_begin > seek->orig_begin)
      || (seek->dir == LIST_BACK && curr_orig_begin < seek->orig_begin))
    return FALSE;

  return TRUE;
}

static void diff_seek(diff_t diff, size_t orig_begin)
{
  diff_block_t curr_block;
  struct diff_seek_traverse seek;
  size_t curr_orig_begin;
  int ret;

  if (list_empty(diff->diff_block_list))
    return;

  seek.orig_begin = orig_begin;

  curr_block = list_curr(diff->diff_block_list);
  diff_block_begin_get(curr_block, &curr_orig_begin, NULL);
  if (orig_begin > curr_orig_begin)
    seek.dir = LIST_FORW;
  else
    seek.dir = LIST_BACK;
  ret = list_traverse(diff->diff_block_list, &seek, (list_traverse_func_t)diff_seek_traverse, LIST_ALTR|LIST_CURR|seek.dir);

  /*
   * in a forward traversal that doesn't hit the end of the list, we
   * need to backtrack one unless if we've hit an exact match.
   */
  curr_block = list_curr(diff->diff_block_list);
  diff_block_begin_get(curr_block, &curr_orig_begin, NULL);
  if (curr_orig_begin > orig_begin
      && ret != LIST_EXTENT)
    list_mvprev(diff->diff_block_list);
}
