/*
 * Copyright (C) 2011 Nathan Phillip Brink <ohnobinki@ohnopublishing.net>
 *
 * This file is a part of difffilter.
 *
 * difffilter is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * difffilter is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with difffilter.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef _DIFFFILTER_DIFF_H
#define _DIFFFILTER_DIFF_H

/**
 * \file
 *
 * \brief An abstraction around a diff as a whole, which consists of
 *   multiple diff_blocks
 */

#include "block.h"

#include <stdio.h>
#include <unistd.h>

struct diff;
typedef struct diff *diff_t;

/**
 * \brief Allocate memory for and initialize a diff.
 *
 * \todo Support parsing and rendering of the timestamps displayed
 * after the --- and +++ lines.
 *
 * \param orig_filename The filename to display on a diff's ``---'' line.
 * \param new_filename The filename to display on a diff's ``+++'' line.
 * \return A new and empty diff.
 */
diff_t diff_init(const char *orig_filename, const char *new_filename);

/**
 * \brief Adds a diff_block to this diff.
 *
 * \param diff The diff.
 * \param diff_block The diff_block to append to this diff. You may not free this pointer yourself; we will free it when you call diff_free().
 * \return 0 on success. 1 on failure.
 */
int diff_add(diff_t diff, diff_block_t diff_block);

enum diff_visit_blocks_func_return
  {
    /**
     * \brief Don't delete the current visited diff_block.
     */
    DIFF_VISIT_BLOCKS_KEEP,
    /**
     * \brief Request that the current visited diff_block be deleted.
     */
    DIFF_VISIT_BLOCKS_DELETE,
  };
typedef enum diff_visit_blocks_func_return (*diff_visit_blocks_func_t)(diff_block_t block, void *data);
/**
 * \brief Visit all diff_blocks in a diff and apply func to each one.
 *
 * \param diff The diff.
 * \param func The function to apply to each diff block.
 * \param data A pointer to pass on to func.
 */
int diff_visit_blocks(diff_t diff, diff_visit_blocks_func_t func, void *data);

/**
 * \brief Render an entire diff.
 *
 * \param diff The diff.
 * \param buf Where to store the newly allocated pointer which is the whole diff.
 * \return The length of buf or -1 on error.
 */
ssize_t diff_render(diff_t diff, char **buf);

/**
 * \brief Writes an entire diff to the given file handle.
 *
 * \param diff The diff.
 * \param out The file handle.
 * \return 0 on success. 1 on failure.
 */
int diff_write_file(diff_t diff, FILE *out);

/**
 * \brief Free a diff and all of its associated diff_blocks.
 *
 * \param diff The diff.
 */
void diff_free(diff_t diff);

#endif /* _DIFFFILTER_DIFF_H */
