/*
 * Copyright (C) 2011 Nathan Phillip Brink <ohnobinki@ohnopublishing.net>
 *
 * This file is a part of difffilter.
 *
 * difffilter is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * difffilter is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with difffilter.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "block.h"
#include "diff.h"
#include "mangle.h"
#include "pattern.h"

#include <errno.h>
#include <list.h> /* http://ohnopub.net/liblist */
#include <stdio.h>
/* malloc(), free() */
#include <stdlib.h>
#include <string.h>
/* getline() on non-GNU systems, http://ohnopub.net/~ohnobinki/libstrl */
#include <strl.h>
#include <tre/regex.h>
#include <unistd.h>

int difffilter(FILE *in, FILE *out, list_t bad_files_list, list_t cvs_keyword_list, short reverse, ssize_t context_lines, short cleanup_allwhitespace, short cleanup_somewhitespace);
int diff_error_input(FILE *in);
int diff_error_wrongchar(const char *file, const char *line);
int main_cleanup(list_t pattern_list, list_t cvs_keyword_list);

int main(int argc, char *argv[])
{
  char opt;
  int noop;
  int ret;

  difffilter_pattern_t difffilter_pattern;
  /*
   * 1 if the patterns read in should be marked as ``accept'' or 0 if
   * they should be marked as ``reject''.
   */
  short accept_pattern;

  /*
   * Whether or not the patch should be reversed, -R.
   */
  short reverse;

  /*
   * Whether or not to try to remove whitespace changes, like diff -w.
   */
  short cleanup_allwhitespace;
  /*
   * like cleanup_allwhitespace but for diff -b.
   */
  short cleanup_somewhitespace;

  /*
   * the number of lines of context to reduce the diff to. -1 for
   * disabled/auto. (in disabled/auto, the number of lines of context
   * are detected and then reduced only for when diff lines are
   * nullified.)
   */
  ssize_t context_lines;

  list_t pattern_list;
  list_t cvs_keyword_list;
  regex_t *preg;

  pattern_list = list_init();
  cvs_keyword_list = list_init();

  cleanup_allwhitespace = 0;
  cleanup_somewhitespace = 0;
  reverse = 0;
  context_lines = -1;

  /* ask getopt() to be verbose */
  opterr = 1;

  noop = 0;
  accept_pattern = 0;
  while( (opt = getopt(argc, argv, "abe:E:hk:K:RU:vwx")) != (char)-1 )
    switch(opt)
      {
      case 'h':
	printf(PACKAGE_STRING " - %1$s [-U <lines>] [-w] [-v] [-h] [-e <BRE>] [-a] [-E <ERE>] [-x] [-e <BRE>] [-k <BRE>]\n\
\n\
Reads unified diff from standard input and writes it to output. Any additional parameters are POSIX Basic Regular Expressions (BRE). Any time a filename from a diff header matches the regular expression, that section of the diff is not written to standard output. This action can be changed, however, with the -a and -x options.\n\
\n\
 -h\n\tShow this help.\n\
 -v\n\tPrint the version and credits information for this program.\n\
\n\
 -e <BRE>\n\tProvide a Basic Regular Expression to match against filenames in the diff.\n\
 -E <ERE>\n\tProvide an Extended Regular Expression to match against filenames in the diff.\n\
 -a\n\tAny -e options following this will denote filename patterns that should be accepted and passed through.\n\
 -x\n\tAny -e options following this will denote filename patterns that should be rejected (excluded) and ignored.\n\
\n\
 -b \n\tAttempt to remove or nullify changes in the amount of whitespace.\n\
 -w \n\tAttempt to remove or nullify the effects of whitespace changes.\n\
 -k <BRE>\n\tA Basic Regular Expression to match against CVS- or SVN-style keywords whose changes should be nullified.\n\
 -K <ERE>\n\tAn Extended Regular Expression to match against CVS- or SVN-style keywords whose changes should be nullified.\n\
 -R\n\tReverse the diff.\n\
 -U <lines>\n\tReduce the number of lines of context to <lines>\n\
",
	       argv[0]);
	noop = 1;
	break;

      case 'v':
	printf(PACKAGE_STRING " <" PACKAGE_URL ">\n\
Copyright (C) 2011 Nathan Phillip Brink <ohnobinki@ohnopublishing.net>\n\
\n\
License GNU Affero General Public License version 3 <http://gnu.org/licenses/>\n\
This is free software: you are free to change and redistribute it.\n\
There is NO WARRANTY, to the extent permitted by law.\n");
	noop = 1;
	break;

      case '?':
	main_cleanup(pattern_list, cvs_keyword_list);
	return 1;

      case 'e':
      case 'E':
	difffilter_pattern = difffilter_pattern_new(optarg, accept_pattern, opt == 'E');
	if(!difffilter_pattern)
	  {
	    main_cleanup(pattern_list, cvs_keyword_list);
	    return 1;
	  }
	list_insert_after(pattern_list, difffilter_pattern, 0);
	break;

      case 'R':
	reverse = 1;
	break;

      case 'a':
	accept_pattern = 1;
	break;

      case 'x':
	accept_pattern = 0;
	break;

      case 'b':
	cleanup_somewhitespace = 1;
	break;

      case 'w':
	cleanup_allwhitespace = 1;
	break;

      case 'k':
      case 'K':
	preg = malloc(sizeof(regex_t));
	ret = regcomp(preg, optarg, (opt == 'K' ? REG_EXTENDED : 0) | REG_NOSUB);
	if(ret)
	  {
	    diff_print_regerror(optarg, (reg_errcode_t)ret, preg);
	    regfree(preg);

	    main_cleanup(pattern_list, cvs_keyword_list);
	    return 1;
	  }
	list_insert_after(cvs_keyword_list, preg, 0);
	break;

      case 'U':
	context_lines = atol(optarg);
	if (context_lines < -1)
	  {
	    fprintf(stderr, "error: -U accepts only integers greater than or equal to -1. You specified %ld.\n", (long int)context_lines);
	    main_cleanup(pattern_list, cvs_keyword_list);
	    return 1;
	  }
	break;

      default:
	fprintf(stderr, "huh?\n");
	break;
      }

  if(noop)
    {
      main_cleanup(pattern_list, cvs_keyword_list);
      return 0;
    }

  /*
   * warn about odd option combinations.
   */
  if (cleanup_allwhitespace && cleanup_somewhitespace)
    fprintf(stderr, "warning: Both -b and -w were specified, which should have the same effect as only specifying -w.\n");

  /*
   * maintain support for difffilter-0.2-style patterns, where -e
   * wasn't available.
   */
  for(; optind < argc; optind ++)
    {
      difffilter_pattern = difffilter_pattern_new(argv[optind], accept_pattern, 0);
      if(!difffilter_pattern)
	{
	  main_cleanup(pattern_list, cvs_keyword_list);
	  return 1;
	}

      list_insert_after(pattern_list, difffilter_pattern, 0);
    }

  ret = difffilter(stdin, stdout, pattern_list, cvs_keyword_list, reverse, context_lines, cleanup_allwhitespace, cleanup_somewhitespace);

  main_cleanup(pattern_list, cvs_keyword_list);
  return ret;
}

void main_preg_free(regex_t *preg)
{
  regfree(preg);
  free(preg);
}

int main_cleanup(list_t pattern_list, list_t cvs_keyword_list)
{
  list_free(pattern_list, (list_dealloc_func_t)&difffilter_pattern_free);
  list_free(cvs_keyword_list, (list_dealloc_func_t)&main_preg_free);

  return 0;
}

struct difffilter_traverse_cvs_keyword_info
{
  short bad_keyword;
  char *keyword;
};

int difffilter_traverse_cvs_keyword_check(struct difffilter_traverse_cvs_keyword_info *info, regex_t *preg)
{
  if(!regexec(preg, info->keyword, 0, NULL, 0))
    {
      info->bad_keyword = 1;

      /* short-curuit the traversal */
      return FALSE;
    }
  return TRUE;
}

struct difffilter_traverse_filecheck_info
{
  short *skip_file;
  char *filename;
};

int difffilter_traverse_filecheck(struct difffilter_traverse_filecheck_info *info, difffilter_pattern_t badfile_pattern)
{
  if(!regexec(&badfile_pattern->preg, info->filename, 0, NULL, 0))
    {
      if(badfile_pattern->accept)
	*info->skip_file = 0;
      else
	*info->skip_file = 1;

      /*
       * short-circuit the traversal. Required to make the
       * badfile_pattern->accept work.
       */
      return FALSE;
    }
  return TRUE;
}

int difffilter(FILE *in, FILE *out, list_t bad_files_list, list_t cvs_keyword_list, short reverse, ssize_t context_lines, short cleanup_allwhitespace, short cleanup_somewhitespace)
{
  ssize_t line_len;
  char *line;
  size_t line_alloc_size;

  struct difffilter_traverse_filecheck_info traverse_info;

  short cleanup_cvs_keywords;
  short skip_file;
  short garbage;
  short in_diff;
  short line_read;
  char *diff_headers[2]; /*< we must cache the diff headers because we have to check our filters against both headers */
  char *filename; /*< The file that the diff header specifies which we're working with */
  char *tmp;

  size_t len;

  unsigned long orig_begin;
  unsigned long orig_numlines;
  unsigned long new_begin;
  unsigned long new_numlines;
  char *block_comment;
  int scanf_matches;

  diff_t diff;
  diff_block_t diff_block;

  traverse_info.skip_file = &skip_file;

  diff = NULL;
  diff_block = NULL;
  line = NULL;
  line_alloc_size = 0;
  line_read = 0;
  line_len = 0; /*< Used to supress a non-applicable warning */
  cleanup_cvs_keywords = list_size(cvs_keyword_list) > 0;

  while(1)
    {
      skip_file = 0;

      /*
	eat garbage lines / header.
	garbage is set to the number of header lines we need to worry about.
       */
      garbage = 2;
      filename = NULL;
      while(garbage > 0)
	{
	  if (!line_read)
	    line_len = getline(&line, &line_alloc_size, in);
	  line_read = 0;

	  if(line_len == -1)
	    {
	      if(feof(in))
		{
		  /* often just a normal EOF */
		  free(line);
		  return 0;
		}
	      else
		{
		  return diff_error_input(in);
		}
	    }

	  /*
	   * Either we have one line of `+++' or two lines, `---' and
	   * then `+++'. Any other sort of header is invalid.
	   */
	  if(line_len > 4
	     && ((garbage == 2 && !strncmp("--- ", line, 4))
		 || !strncmp("+++ ", line, 4)))
	    {
	      /* grasp the filename */
	      tmp = &line[3];
	      while(*tmp == ' ' || *tmp == '\t')
		tmp ++;
	      free(filename);
	      filename = strdup(tmp);

	      /* delimit the end of the filename */
	      tmp = filename;
	      while(*tmp != '\0' && *tmp != ' ' && *tmp != '\t' && *tmp != '\n')
		tmp ++;

	      if(!*filename)
		{
		  fprintf(stderr, "Expecting filename, got nothing: \"%s\"\n", line);

		  /* Reset input heading counter counter and start looking for other headers. */
		  while (garbage <= 2)
		    {
		      free(diff_headers[garbage]);
		      garbage ++;
		    }
		}
	      else
		{
		  /* Insert the null terminator after the end of the scanned-in filename */
		  *tmp = '\0';

		  /* check if the filename matches */
		  traverse_info.filename = filename;
		  list_traverse(bad_files_list, &traverse_info, (list_traverse_func_t)&difffilter_traverse_filecheck, LIST_FORW | LIST_FRNT | LIST_SAVE);

		  garbage --;
		  diff_headers[garbage] = strdup(line);
		  /* cut the newline from the diff_header line */
		  diff_headers[garbage][line_len - 1] = '\0';

		  /*
		   * Support diffs with ``short headers'', which look like:
		   *
		   * +++ b/filename
		   * @@ -1,2 +1,2 @@
		   *
		   * These diffs don't have a `---' line and thus we
		   * should handle that correctly. We detect this by
		   * garbage being positive (we're expecting to read
		   * another header in) but this current header
		   * already being `+' (meaning that we're arrived at
		   * the end of the headers already).
		   */
		  if (garbage && line[0] == '+')
		    {
		      /* Same thing as assigning 0 to garbage... */
		      garbage --;
		      diff_headers[garbage] = strdup(diff_headers[garbage + 1]);
		      memcpy(diff_headers[garbage + 1], "---", 3);
		    }
		}
	    }
	  else
	    /* Do not allow there to be any lines between the `---' line and the `+++' line: */
	    if (garbage < 2)
	    {
	      /* cut newline */
	      line[line_len - 1] = '\0';
	      fprintf(stderr, "%s: Badly formed diff: expecting \"+++ <filename>\", got \"%s\"\n", filename ? filename : "<unknown>", line);
	      free(filename);
	      free(line);
	      for (; garbage < 2; garbage ++)
		free(diff_headers[garbage]);
	      return 1;
	    }

	  /**
	     If we haven't hit the diff header itself yet, we still passthrough.
	     This gets us, e.g., mercurial's ``hg export'' headers, etc.
	  */
	  if(!filename) /* same thing as checking garbage >= 2 */
	      fputs(line, out);
	} /* while(garbage > 0) */

      /**
       * regurgitate the unified diff headers now that we know whether
       * or not to skip_file ;-).
       *
       * By now we know that the diff_headers array has two valid
       * members.
       */
      if(!skip_file)
	{
	  /**
	   * Reverse the diff headers if we are to reverse the diff as
	   * a whole.
	   */
	  if(reverse)
	    {
	      tmp = diff_headers[1];
	      diff_headers[1] = diff_headers[0];
	      diff_headers[0] = tmp;

	      memcpy(diff_headers[0], "---", 3);
	      memcpy(diff_headers[1], "+++", 3);
	    }

	  /*
	   * cheat -- we pass the timestamps and stuff with the
	   * filename for now instead of understanding and
	   * reconstructing the timestamp. In fact, this behavior is
	   * probably preferable as the less we change the diff, the
	   * better.
	   */
	  diff = diff_init(&diff_headers[1][4], &diff_headers[0][4]);
	}

      /**
	 Parse diff magic headers, etc.
      */
      in_diff = 1;
      orig_begin = 0;
      orig_numlines = 0; /*< The number of lines left in the original */
      new_begin = 0;
      new_numlines = 0; /*< The number of lines left in the new version */
      
      while(in_diff > 0)
	{
	  /*
	   * Must always read a new line in even if we've finished a
	   * diff block because we don't know if we'll get another
	   * diff block (`@@ ....') or something else yet. But if this
	   * line we read ends up not being another diff block, it
	   * should be kept as it may be another diff file's header
	   * (line_read = 1).
	   */
	  line_len = getline(&line, &line_alloc_size, in);
	  line_read = 1;

	  if(orig_numlines == 0
	     && new_numlines == 0)
	    in_diff = 0;

	  /* update the counters based on the current line */
	  if(in_diff)
	    {
	      /*
		we must check for I/O errors _after_ in_diff is updated.
		Otherwise, we'll croak if the EOF happens to be immediately
		after the diff is over ;-)
	       */
	      if(line_len == -1)
		return diff_error_input(in);

	      switch(line[0])
		{
		case ' ':
		  /* check if we're trying to go below zero: */
		  if(orig_numlines == 0
		     || new_numlines == 0)
		    return diff_error_wrongchar(filename, line);
		  orig_numlines --;
		  new_numlines --;
		  break;

		case '+':
		  if(new_numlines == 0)
		    return diff_error_wrongchar(filename, line);
		  new_numlines --;

		  if(reverse)
		    line[0] = '-';
		  break;

		case '-':
		  if(orig_numlines == 0)
		    return diff_error_wrongchar(filename, line);
		  orig_numlines --;

		  if(reverse)
		    line[0] = '+';
		  break;

		  /*
		    handle ``\ No newline at end of file'', but doesn't detect
		    multiple occurences of ``\'' (which would be incorrect).
		   */
		case '\\':
		  if(orig_numlines
		     && new_numlines)
		    return diff_error_wrongchar(filename, line);
		  break;

		default:
		  fprintf(stderr, "Unrecognized diff character '%c' for file \"%s\" in \"%s\"\n",
			  line[0], filename, line);
		  return 1;
		}
	    }

	  /*
	   * parse a line into the diff_block. Only when we're in_diff.
	   */
	  if(in_diff
	     && !skip_file
	     && line_len != -1)
	    {
	      if(diff_block_interpret(diff_block, line))
		{
		  fprintf(stderr, "error: diff_block_interpret() returned error for line: ``%s''\n",
			  line);
		  return 1;
		}
	    }

	  /*
	   * Once we've gotten through a diff block, pop it into out
	   * diff handle.
	   */
	  if(!in_diff
	     && !skip_file
	     && diff_block)
	    {
	      if(diff_add(diff, diff_block))
		{
		  fprintf(stderr, "error: diff_add() returned error\n");
		  return 1;
		}
	      /*
	       * now that we've added diff_block to diff, we may not
	       * manage its memory anymore.
	       */
	      diff_block = NULL;
	    }

	  /**
	     Once in_diff drops to zero, that means we have to look for and parse another
	     magic diff line:
	     @@ -0,0 +1,29 @@
	     
	     That means the original file's has zero lines of diff and the new
	     file has 29 lines of diff starting at line 1. Thus we must skip
	     zero `-' lines, zero combined ` ' lines, and 29 `+' lines.

	     The following format is also valid:
	     @@ -1 +1 @@
	     It means:
	     @@ -1,1 +1,1 @@

	     OR, in_diff == 0 means that we're done with diffs for this file.
	   */
	  if(!in_diff
	     && line_len != -1)
	    {
	      /* check for inconsistent state */
	      if(diff_block)
		{
		  fprintf(stderr, "error: inconsistent state: diff_block is non-NULL\n");
		  return 1;
		}

	      if(line_len > 2
		 && !strncmp("@@ ", line, 3))
		{
		  /*
		    Use in_diff to track if we came across a valid magic diff header or not.
		    (it's already set to 0 atm)
		   */

		  /* parse diff magic header */
		  scanf_matches = sscanf(line, "@@ -%lu,%lu +%lu,%lu @@",
					 &orig_begin, &orig_numlines,
					 &new_begin, &new_numlines);
		  if(scanf_matches == 4)
		    in_diff = 1;

		  if(!in_diff)
		    {
		      new_numlines = 1;
		      scanf_matches = sscanf(line, "@@ -%lu,%lu +%lu @@",
					     &orig_begin, &orig_numlines,
					     &new_begin);
		      if(scanf_matches == 3)
			in_diff = 1;
		    }

		  if(!in_diff)
		    {
		      orig_numlines = 1;
		      scanf_matches = sscanf(line, "@@ -%lu +%lu,%lu @@",
					     &orig_begin,
					     &new_begin, &new_numlines);
		      if(scanf_matches == 3)
			in_diff = 1;
		    }

		  /*
		   * This entry must be last because scanf() will
		   * accept "23,43 @@" for the format string "%lu
		   * @@". I don't think I understand completely the
		   * scanf() API.
		   */
		  if(!in_diff)
		    {
		      orig_numlines = 1;
		      new_numlines = 1;
		      scanf_matches = sscanf(line, "@@ -%lu +%lu @@",
					     &orig_begin,
					     &new_begin);
		      if(scanf_matches == 2)
			in_diff = 1;
		    }

		  /*
		   * Make sure that the delimiter found by getline()
		   * doesn't get into the block_comment.
		   */
		  line[line_len - 1] = '\0';

		  block_comment = strstr(&line[2], "@@") + 2;
		  if (*block_comment && strchr(" \t", *block_comment))
		    block_comment ++;
		  if (!*block_comment)
		    block_comment = NULL;

		  /* bad magic diff header */
		  if(!in_diff)
		    {
		      fprintf(stderr, "%s: I tried to parse \"%s\" as a magic diff header and that failed :-/\n",
			      filename, line);
		      return 1;
		    }

		  /*
		   * we are in a diff. If reversing the diff, this is
		   * the time to touch the magic header. We might as
		   * well just format our own, because cutting and
		   * pasting from the original would make the code
		   * confusing. ;-)
		   */
		  if(reverse)
		    {
		      /*
		       * @@ -1 +1 @@ should only expand to @@ -1,1 +1,1 @@
		       */
		      len = strlen(line) + 2 + 2 + 1;
		      if(line_alloc_size < len)
			{
			  free(line);
			  line = malloc(len);
			  line_alloc_size = len;
			}
		      if(snprintf(line, len, "@@ -%lu,%lu +%lu,%lu @@%s",
				  new_begin, new_numlines,
				  orig_begin, orig_numlines,
				  block_comment) > (len - 1))
			{
			  fprintf(stderr, "%s: Unable to rewrite magic diff header so as to reverse the diff. snprintf() wanted to use more than %zu bytes.\n",
				  filename, len);
			  return 1;
			}
		    }

		  /*
		   * start a new diff_block
		   */
		  if(!skip_file)
		    {
		      /*
		       * See comments in the block_render() function
		       * regarding orig_begin_translated and its
		       * friend. Patch utilities will refer to a
		       * nonexistent line by referring to the previous
		       * line and by setting _numlines to 0. Our logic
		       * likes to pretend that the nonexistent line
		       * does exist but that it has a depth of zero.
		       */
		      if (!new_numlines)
			new_begin ++;
		      if (!orig_numlines)
			orig_begin ++;

		      if(reverse)
			diff_block = diff_block_init(new_begin, orig_begin);
		      else
			diff_block = diff_block_init(orig_begin, new_begin);
		      diff_block_comment_set(diff_block, block_comment);
		    }
		}
	    }
	} /* while(in_diff > 0) */

      if (!skip_file)
	{
	  /*
	   * Perform post-processing.
	   */
	  if (cleanup_cvs_keywords)
	    mangle_killchange(diff, (mangle_killchange_filter_func_t)&mangle_killchange_cvs_keywords_filter, cvs_keyword_list);
	  if (cleanup_allwhitespace)
	    mangle_killchange(diff, &mangle_killchange_whitespace_filter, NULL);
	  if (cleanup_somewhitespace)
	    mangle_killchange(diff, &mangle_killchange_whitespaceamount_filter, NULL);
	  if (context_lines != -1)
	    mangle_condense(diff, context_lines);

	  /*
	   * write the buffered diff to standard out.
	   */
	  diff_write_file(diff, out);
	  diff_free(diff);
	  diff = NULL;
	}

    } /* while(1) */
  free(filename);
  free(line);

  return 0;
}

int diff_error_wrongchar(const char *file, const char *line)
{
  fprintf(stderr, "%s:Badly formed diff: unexpected diff character '%c', the magic diff header is probably messed up near \"%s\"\n", file, line[0], line);
  return 1;  
}

int diff_error_input(FILE *in)
{
  int errortype;
  static char *error_types[] =
    {
      "unexpected end of file",
      "generic input error",
      "unexpected aliens invading from outer space",
    };

  errortype = 2;

  if(feof(in))
    errortype = 0;
  if(ferror(in))
    errortype = 1;

  fprintf(stderr, "Reading input interrupted: %s%s\n",
	  error_types[errortype], errortype == 1 ? strerror(errno) : "");
  return 1;
}

