/*
 * Copyright (C) 2011 Nathan Phillip Brink <ohnobinki@ohnopublishing.net>
 *
 * This file is a part of difffilter.
 *
 * difffilter is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * difffilter is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with difffilter.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "mangle.h"

#include <list.h>
#include <stdlib.h>
#include <string.h>
#include <tre/regex.h>

/**
 * \brief
 *   Split the a diff block's diff_block_line list at list_curr().
 *
 * A helper for mangle_condense().
 *
 * The current element of the list will be the first element of the
 * newly generated block. The newly generated block is added to the
 * overflow_blocks list and returned.
 *
 * \param lines The diff_block_line list to split off another block
 *   from.
 * \param overflow_blocks The list into which newly created
 *   diff_blocks should be stored.
 * \param orig_begin The origin_begin numeral of the magic diff
 *   header.
 * \param new_begin The new_begin numeral of the magic diff header.
 *
 * \return The newly created block (the second half of the split).
 */
diff_block_t mangle_condense_split(list_t lines, list_t overflow_blocks, size_t orig_begin, size_t new_begin)
{
  diff_block_t new_diff_block;
  list_t new_block_lines;
  void *curr_line;
  void *last_line;

  new_diff_block = diff_block_init(orig_begin, new_begin);
  new_block_lines = diff_block_lines(new_diff_block);
  curr_line = list_curr(lines);
  last_line = NULL;

  /* remove lines until we hit the current one */
  while (curr_line != last_line)
    {
      last_line = list_remove_rear(lines);
      list_insert_before(new_block_lines, last_line, 0);
    }

  list_insert_after(overflow_blocks, new_diff_block, 0);

  return new_diff_block;
}

enum diff_visit_blocks_func_return mangle_condense_visit(diff_block_t orig_block, struct mangle_condense *condense)
{
  diff_block_t block;

  list_t lines;
  struct diff_line *line;

  size_t num_noop;
  /* mark that we're neither at the end or beginning of the block */
  short middle;

  /* we need to remember what lines we're on */
  size_t orig_line;
  size_t new_line;

  block = orig_block;
  lines = diff_block_lines(block);
  diff_block_begin_get(block, &orig_line, &new_line);

  num_noop = 0;
  middle = 0;

  list_mvfront(lines);
  line = list_curr(lines);
  while (line)
    {
      if (line->type != DIFF_LINE_NOOP)
	{
	  /* backtrack and reduce lines if necessary */

	  if (!middle)
	    {
	      middle = 1;

	      /* case 1: we're removing lines from the beginning of a block */
	      if (num_noop > condense->context)
		{
		  /* back out */
		  diff_block_begin_get(block, &orig_line, &new_line);
		  list_mvfront(lines);

		  for (; num_noop > condense->context; num_noop --)
		    {
		      diff_line_free(list_remove_curr(lines));
		      orig_line ++;
		      new_line ++;
		    }
		  diff_block_begin_set(block, orig_line, new_line);
		}
	    }

	  /* case 2: we're splitting a block into two blocks (or cutting the end off) */
	  if (middle && num_noop > condense->context * 2)
	    {
	      while (num_noop > condense->context)
		{
		  list_mvprev(lines);
		  num_noop --;
		  orig_line --;
		  new_line --;
		}

	      block = mangle_condense_split(lines, condense->new_blocks, orig_line, new_line);
	      lines = diff_block_lines(block);
	      diff_block_begin_get(block, &orig_line, &new_line);

	      list_mvfront(lines);

	      middle = 0;
	    }

	  /* reset state */
	  num_noop = 0;

	  line = list_curr(lines);
	}

      /*
       * after the above rearranging, we may end up on a noop line
       * still, so need to recheck
       */
      if (line->type == DIFF_LINE_NOOP)
	num_noop ++;

      if (list_mvnext(lines))
	{
	  /*
	   * \todo incrementing based on the line type is also used in
	   *   difffilter.c, perhaps this can befactored into a
	   *   function.
	   */
	  switch (line->type)
	    {
	    case DIFF_LINE_NOOP:
	      new_line ++;
	      orig_line ++;
	      break;

	    case DIFF_LINE_MINUS:
	      orig_line ++;
	      break;

	    case DIFF_LINE_PLUS:
	      new_line ++;
	      break;
	    }

	  /*
	   * The above new_line and orig_line incrementations must be
	   * based on the previous line in the diff block. This is
	   * because only the previous line defines whether or not
	   * this line's new_line or orig_line are greater than the
	   * values for the previous line.
	   */
	  line = list_curr(lines);
	}
      else
	line = NULL;
    }

  /*
   * if we've never encountered the middle of the diff, then this diff
   * only has noop lines in it
   */
  if (!middle)
    {
      if (block == orig_block)
	return DIFF_VISIT_BLOCKS_DELETE;
      else
	/*
	 * we are dealing with a block newly created by
	 * mangle_condense_split()
	 */
	diff_block_free(list_remove_curr(condense->new_blocks));
    }

  /* trim lines off of the end of a block */
  while (num_noop > condense->context)
    {
      diff_line_free(list_remove_curr(lines));
      num_noop --;
    }

  return DIFF_VISIT_BLOCKS_KEEP;
}

static int mangle_condense_diff_add(diff_t diff, diff_block_t block)
{
  diff_add(diff, block);
  return TRUE;
}

int mangle_condense(diff_t diff, size_t context_lines)
{
  struct mangle_condense condense;

  int ret;

  condense.context = context_lines;
  condense.new_blocks = list_init();
  if (!condense.new_blocks)
    return 1;

  ret = diff_visit_blocks(diff, (diff_visit_blocks_func_t)&mangle_condense_visit, &condense);
  list_traverse(condense.new_blocks, diff, (list_traverse_func_t)&mangle_condense_diff_add, LIST_FORW|LIST_FRNT|LIST_SAVE);
  list_free(condense.new_blocks, NULL);

  return ret;
}

char *mangle_killchange_whitespace_filter(const char *in, void *data)
{
  size_t in_pos;
  size_t out_pos;
  char *out;

  out = malloc(strlen(in) + 1);
  if (!out)
    return NULL;

  in_pos = 0;
  out_pos = 0;
  while (in[in_pos])
    {
      if (!strchr(" \t\r\n", in[in_pos]))
	{
	  out[out_pos] = in[in_pos];
	  out_pos ++;
	}
      in_pos ++;
    }
  out[out_pos] = '\0';

  return out;
}

/**
 * \brief
 *   Remove items that diff -b would ignore from a string.
 *
 * A helper for mangle_killchange_visit(). This is different from
 * mangle_killchange_whitespace_filter() because this declares a
 * distinction between no whitespace and some whitespace while
 * ignoring differences in the amount of whitespace.
 *
 * \param in
 *   The string from which to make a copy stripped of whitespace.
 * \return
 *   A new string with the appropriate whitespace chars removed or
 *   NULL on error.
 */
char *mangle_killchange_whitespaceamount_filter(const char *in, void *data)
{
  size_t in_pos;
  size_t out_pos;
  short saw_space;
  char *out;

  out = malloc(strlen(in) + 1);
  if (!out)
    return NULL;

  saw_space = 0;
  in_pos = 0;
  out_pos = 0;
  while (in[in_pos])
    {
      if (!strchr(" \t\r\n", in[in_pos]))
	{
	  out[out_pos] = in[in_pos];
	  out_pos ++;
	  saw_space = 0;
	}
      else
	{
	  if (!saw_space)
	    {
	      out[out_pos] = ' ';
	      out_pos ++;
	      saw_space = 1;
	    }
	}
      in_pos ++;
    }
  out[out_pos] = '\0';

  return out;
}

static int mangle_killline_cvs_keywords_filter_traverse(char *submatch, regex_t *keyword_preg)
{
  if (regexec(keyword_preg, submatch, 0, NULL, REG_NOSUB))
    /* no match */
    return TRUE;
  /* match, stop traversal early. */
  return FALSE;
}

char *mangle_killchange_cvs_keywords_filter(const char *in, list_t cvs_keyword_list)
{
  static regex_t cvs_keyword_preg;
  static int cvs_keyword_preg_init = 0;

  regmatch_t cvs_keyword_matches[2];
  char *submatch;
  size_t submatch_len;

  char *out;
  char *cur;

  /* first/last time init/deinit-ialization */
  if (!in)
    {
      if (cvs_keyword_preg_init)
	regfree(&cvs_keyword_preg);
      cvs_keyword_preg_init = 0;
      /* return now, quick, before we reinitialize cvs_keyword_preg_init ;-) */
      return NULL;
    }
  if (!cvs_keyword_preg_init)
    {
      /** The first subgroup will be matched against cvs_keyword_list */
      if (regcomp(&cvs_keyword_preg, "\\$([^: $]+)(:[^$]*)?\\$", REG_EXTENDED))
	return NULL;
      cvs_keyword_preg_init = 1;
    }

  out = strdup(in);
  if (!out)
    return NULL;

  cur = out;
  while (*cur && (cur = strchr(cur, '$')))
    {
      if (regexec(&cvs_keyword_preg, cur, 2, cvs_keyword_matches, 0)
	  || cvs_keyword_matches[1].rm_so == -1)
	{
	  /*
	   * Match failed or some undetected error or the first
	   * subexpression didn't get any match.
	   */
	  cur ++;
	  continue;
	}

      /*
       * Check if the matched substring for that last regex matches
       * one of the passed in CVS keywords regexes.
       */
      submatch_len = cvs_keyword_matches[1].rm_eo - cvs_keyword_matches[1].rm_so;
      submatch = malloc(submatch_len + 1);
      if (!submatch)
	{
	  free(out);
	  return NULL;
	}
      memcpy(submatch, cur + cvs_keyword_matches[1].rm_so, submatch_len);
      submatch[submatch_len] = '\0';

      /*
       * If the list is empty, returns LIST_EMPTY. If no match,
       * returns LIST_EXTENT. Thus, we know there's a match with
       * LIST_OK.
       */
      if (list_traverse(cvs_keyword_list, submatch, (list_traverse_func_t)&mangle_killline_cvs_keywords_filter_traverse, LIST_FORW|LIST_FRNT|LIST_SAVE) == LIST_OK)
	{
	  /* match, act as if this keyword doesn't exist */
	  memmove(cur + cvs_keyword_matches[0].rm_so,
		  cur + cvs_keyword_matches[0].rm_eo,
		  strlen(cur + cvs_keyword_matches[0].rm_eo) + 1);
	}
      else
	/* no matches, keep this cvs keyword and move on */
	cur += cvs_keyword_matches[0].rm_eo;
      free(submatch);
    }

  return out;
}

struct mangle_killchange_metaline
{
  struct diff_line *line;
  char *filtered_line;
  /*
   * help determine the ordering of lines when piecing them together
   */
  size_t serial;
};

/**
 * \brief
 *   Allocates and initializes a struct mangle_killchange_metaline.
 */
static struct mangle_killchange_metaline *
mangle_killchange_metaline_init(struct diff_line *line,
				    char *filtered_line,
				    size_t serial)
{
  struct mangle_killchange_metaline *metaline;

  metaline = malloc(sizeof(struct mangle_killchange_metaline));
  if (!metaline)
    return NULL;

  metaline->line = line;
  metaline->filtered_line = filtered_line;
  metaline->serial = serial;

  return metaline;
}

/**
 * \brief
 *   Replace a common area between left and right with a NOOP section.
 *
 * Searches out the largest set of common filtered lines between the
 * left and right set of lines. If a common area is found, the lines
 * before the common area and after the common area shall be
 * recursively checked for common areas. Common areas are replaced
 * with NOOP sections.
 *
 * The purpose of this function is to facilitate a sort of post facto
 * ``diff -b'' and CVS keyword removal. The same perhaps should
 * instead be achieved using a more sophisticated diff algorithm
 * through shelling out to diff(1) or using xdiff or whatnot.
 *
 * This function, however, will cover the simple cases well enough.
 *
 * \param left
 *   The list of struct mangle_killchange_metaline* which are
 *   DIFF_LINE_MINUS. After this function is completed, list_free()
 *   will already have been called on left so you may not use the list
 *   you pass into this function again.
 * \param right
 *   The list of struct mangle_killchange_metaline* which are
 *   DIFF_LINE_PLUS. After this function is completed, list_free()
 *   will already have been called on left so you may not use the list
 *   you pass into this function again.
 * \param new_lines
 *   This should be a list of struct diff_line* to which we should
 *   append resulting new diff lines. We will not modify the position
 *   of the current element of this list, so if you want us to insert
 *   items into the middle of an existing list you may do so.
 */
void mangle_killwhitespace_matchlines(list_t left, list_t right, list_t new_lines)
{
  list_t sub_left;
  list_t sub_right;

  /*
   * pointer to the element in the left and right with the longest common area.
   */
  void *best_left;
  void *best_right;
  size_t best_len;

  /*
   * pointer to the current set of common lines being checked.
   */
  struct mangle_killchange_metaline *curr_left;
  struct mangle_killchange_metaline *curr_right;
  size_t curr_len;

  /*
   * lookahead pointers
   */
  struct mangle_killchange_metaline *ptr_left;
  struct mangle_killchange_metaline *ptr_right;

  /*
   * cache the pointers to the end of the lists
   */ 
  struct mangle_killchange_metaline *last_left;
  struct mangle_killchange_metaline *last_right;

  best_left = NULL;
  best_right = NULL;
  best_len = 0;

  last_left = list_rear(left);
  last_right = list_rear(right);

  /*
   * find the longest match
   */
  list_mvfront(left);
  curr_left = list_curr(left);
  while (curr_left)
    {
      list_mvfront(right);
      curr_right = list_curr(right);

      while (curr_right)
	{
	  if (!strcmp(curr_left->filtered_line, curr_right->filtered_line))
	    {
	      curr_len = 0;
	      ptr_left = curr_left;
	      ptr_right = curr_right;
	      while (ptr_left && ptr_right
		     && !strcmp(ptr_left->filtered_line, ptr_right->filtered_line))
		{
		  curr_len ++;

		  /* advance and check if we were on a last line */
		  if (ptr_left == last_left
		      || ptr_right == last_right)
		    {
		      ptr_left = NULL;
		      ptr_right = NULL;
		    }
		  else
		    {
		      list_mvnext(left);
		      list_mvnext(right);
		      ptr_left = list_curr(left);
		      ptr_right = list_curr(right);
		    }
		}
	      if (curr_len > best_len)
		{
		  best_len = curr_len;
		  best_left = curr_left;
		  best_right = curr_right;
		}

	      /* rewind and continue searching for the largest common area */
	      while (list_curr(left) != curr_left)
		list_mvprev(left);
	      while (list_curr(right) != curr_right)
		list_mvprev(right);
	    }

	  /* incr and check if at end of list */
	  if (curr_right == last_right)
	    curr_right = NULL;
	  else
	    {
	      list_mvnext(right);
	      curr_right = list_curr(right);
	    }
	}

      /* incr and check if at end of list */
      if (curr_left == last_left)
	curr_left = NULL;
      else
	{
	  list_mvnext(left);
	  curr_left = list_curr(left);
	}
    }

  /*
   * the next two branches need left and right set to the front.
   */
  list_mvfront(left);
  list_mvfront(right);

  ptr_left = list_curr(left);
  ptr_right = list_curr(right);

  /*
   * There's no way to reduce the current block.
   *
   * This is also something known as the base case.
   */
  if (!best_len)
    {
      while (ptr_left || ptr_right)
	{
	  if (ptr_left)
	    {
	      if (!ptr_right || ptr_left->serial < ptr_right->serial)
		{
		  /*
		   * list_remove_curr() will set the current element to the next element
		   */
		  list_remove_curr(left);
		  list_insert_after(new_lines, ptr_left->line, 0);

		  free(ptr_left->filtered_line);
		  free(ptr_left);

		  ptr_left = list_curr(left);
		}
	    }
	  if (ptr_right)
	    {
	      if (!ptr_left || ptr_right->serial < ptr_left->serial)
		{
		  list_remove_curr(right);
		  list_insert_after(new_lines, ptr_right->line, 0);

		  free(ptr_right->filtered_line);
		  free(ptr_right);

		  ptr_right = list_curr(right);
		}
	    }
	}

      /*
       * both of these lists should be empty by now.
       */
      if (!list_empty(left))
	fprintf(stderr, "%s:%d: internal state inconsistent: left list still has items in it.\n",
		__FILE__, __LINE__);
      list_free(left, LIST_NODEALLOC);
      if (!list_empty(right))
	fprintf(stderr, "%s:%d: internal state inconsistent: right list still has items in it.\n",
		__FILE__, __LINE__);
      list_free(right, LIST_NODEALLOC);

      return;
    }

  /*
   * now that we know the position of the largest common lines, split
   * up the problem halfway. We perform our left recursion right here,
   * mirroring an in-order traversal's approach to tree traversal.
   */
  sub_left = list_init();
  sub_right = list_init();

  while (ptr_left != best_left && ptr_left)
    {
      list_insert_after(sub_left, list_remove_curr(left), 0);
      ptr_left = list_curr(left);
    }
  while (ptr_right != best_right && ptr_right)
    {
      list_insert_after(sub_right, list_remove_curr(right), 0);
      ptr_right = list_curr(right);
    }
  if (!ptr_left || !ptr_right)
    {
      fprintf(stderr, "error: internal state inconsistent: %s%s%s\n",
	      (!ptr_left) ? "ptr_left == NULL" : "",
	      (!ptr_left && !ptr_right) ? " and " : "",
	      (!ptr_right) ? "ptr_right == NULL" : "");
    }

  /*
   * this call will result in the proper list_free() calls on sub_left
   * and sub_right, so we don't worry about that.
   */
  mangle_killwhitespace_matchlines(sub_left, sub_right, new_lines);

  /*
   * Set all of the common lines to be NOOP and ignore the right-side
   * lines because those contain the whitespace/CSV-keywords or any
   * other filtered out changes that we're trying to purge from the
   * diff.
   */
  while(best_len)
    {
      ptr_left = list_remove_curr(left);
      ptr_right = list_remove_curr(right);

      ptr_left->line->type = DIFF_LINE_NOOP;
      list_insert_after(new_lines, ptr_left->line, 0);

      free(ptr_left->filtered_line);
      free(ptr_left);

      diff_line_free(ptr_right->line);
      free(ptr_right->filtered_line);
      free(ptr_right);

      best_len --;
    }

  /*
   * right side of the recursion.
   */
  mangle_killwhitespace_matchlines(left, right, new_lines);

  /*
   * The right side will end in the base case, so we're done at this point :-)
   */
}

/*
 * Now that I've started thinking about it, this is actually a
 * somewhat difficult problem to solve... especiall when a diff block
 * has more than one each of DIFF_LINE_PLUS and DIFF_LINE_MINUS.
 *
 *
 * To solve this problem, one has to do a ``side-by-side'' comparison
 * of sorts, but our diff_blocks are flat/linear.
 *
 * To be most effective, the solution should first find the biggest
 * areas of similarity within a single diff block so that a one or two
 * -line apparent whitespace change doesn't prevent the bigger areas
 * of ``whitespace-only changes''.
 *
 * I propose that there be determined a simple filter which strips all
 * whitespace from lines and that a new block be formed by applying
 * this filter to each line of the current block. Then a table should
 * be generated which determines which DIFF_LINE_PLUS lines match a
 * given DIFF_LINE_MINUS line. Then this table will be searched for
 * the biggest area of similarity, condense that, and then repeat
 * until all such blocks are NOOPed.
 *
 * One ``easy/hacky'' solution would be to reconstruct the original
 * and new portions of the diff, write them out to two files, and
 * shell out to ``diff -u <x> -b file.a file.b'' ...
 */
enum diff_visit_blocks_func_return mangle_killchange_visit(diff_block_t block, struct mangle_killchange *killchange)
{
  /*
   * These hold the type struct mangle_killchange_metaline*
   *
   * NO. they have to hold structs which have the mangled form and a
   * reference back to the original struct *line so that these lines
   * may be removed from the list and mangled to DIFF_LINE_NOOP
   */
  list_t left_lines;
  list_t right_lines;

  list_t lines;
  struct diff_line *line;
  void *line_rear;

  /*
   * used for interaction between future list_remove_curr() and list_mvnext() calls.
   */
  short removed_line;

  /* a way to keep the lines ordered. */
  size_t serial;
  struct mangle_killchange_metaline *metaline;

  left_lines = NULL;
  right_lines = NULL;

  lines = diff_block_lines(block);

  list_mvfront(lines);
  if (list_empty(lines))
    return DIFF_VISIT_BLOCKS_KEEP;
  /* a dummy element for a special case in case DIFF_LINE_NOOP */
  list_insert_before(lines, 0, 0);
  list_mvnext(lines);

  serial = 0;
  line = list_curr(lines);
  line_rear = list_rear(lines);
  while (line)
    {
      removed_line = 0;
      switch (line->type)
	{
	case DIFF_LINE_NOOP:
	  if (left_lines || right_lines)
	    {
	      /*
	       * We have come to the end of some differences. Perform
	       * the search for equal lines among the lines stripped
	       * of whitespace.
	       */
	      if (!left_lines)
		left_lines = list_init();
	      if (!right_lines)
		right_lines = list_init();

	      /*
	       * back out one line. mangle_killwhitespace_matchlines()
	       * uses list_insert_after() on our list, so we backup
	       * before calling and then jump back ahead.
	       *
	       * special case: If there were no DIFF_LINE_NOOP lines
	       * before the first DIFF_LINE_PLUS or DIFF_LINE_MINUS
	       * line, then backing up here would not cause the list's
	       * current element to be moved back as the current
	       * element is now the first element of the list.
	       */
	      list_mvprev(lines);
	      mangle_killwhitespace_matchlines(left_lines, right_lines, lines);
	      list_mvnext(lines);
	      removed_line = 1;

	      /* mangle_killwhitespace() calls list_free() for us: */
	      left_lines = NULL;
	      right_lines = NULL;
	    }

	  break;

	case DIFF_LINE_MINUS:
	  if (!left_lines)
	    left_lines = list_init();
	  metaline = mangle_killchange_metaline_init(line,
						     (*killchange->filter_func)(line->line, killchange->filter_func_data),
						     serial);
	  /*
	   * we're passing the memory management of this line off onto
	   * mangle_killwhitespace_match()
	   */
	  list_remove_curr(lines);
	  removed_line = 1;
	  list_insert_after(left_lines, metaline, 0);
	  break;
	case DIFF_LINE_PLUS:
	  if (!right_lines)
	    right_lines = list_init();
	  metaline = mangle_killchange_metaline_init(line,
						     (*killchange->filter_func)(line->line, killchange->filter_func_data),
						     serial);
	  list_remove_curr(lines);
	  removed_line = 1;
	  list_insert_after(right_lines, metaline, 0);
	  break;
	}

      if (line == line_rear)
	/* we're done */
	line = NULL;
      else
	{
	  serial ++;
	  if (!removed_line)
	    list_mvnext(lines);
	  line = list_curr(lines);
	}
    }


  /*
   * remove the dummy first element we place in the list. We don't
   * need this workaround anymore, as using list_mvprev() below makes
   * no sense since we know we've essentially ``passed'' the end of
   * the list of lines.
   */
  list_remove_front(lines);

  /*
   * handle when a diff block doesn't end with at least one
   * DIFF_LINE_NOOP line.
   */
  if (left_lines || right_lines)
    {
      if (!left_lines)
	left_lines = list_init();
      if (!right_lines)
	right_lines = list_init();

      /*
       * We are past the last line. Thus, we shouldn't do the ``back
       * up one line'' thing here.
       */
      mangle_killwhitespace_matchlines(left_lines, right_lines, lines);
      /* 
       * left_lines and right_lines are free()ed by
       * mangle_killwhitespace_matchlines()
       */
    }

  return DIFF_VISIT_BLOCKS_KEEP;
}

int mangle_killchange(diff_t diff, mangle_killchange_filter_func_t filter_func, void *filter_func_data)
{
  struct mangle_killchange killchange;

  killchange.filter_func = filter_func;
  killchange.filter_func_data = filter_func_data;

  return diff_visit_blocks(diff, (diff_visit_blocks_func_t)&mangle_killchange_visit, &killchange);
}
