/*
 * Copyright (C) 2011 Nathan Phillip Brink <ohnobinki@ohnopublishing.net>
 *
 * This file is a part of difffilter.
 *
 * difffilter is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * difffilter is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with difffilter.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef _DIFFFILTER_MANGLE_H
#define _DIFFFILTER_MANGLE_H

/**
 * \file
 *   Manglers, processors, and filters using the diff.h API.
 */

#include "diff.h"

#include <list.h>

/**
 * \brief
 *  Options for mangle_condense().
 *
 * \see mangle_condense()
 */
struct mangle_condense
{
  /**
   * \brief
   *   Number of lines of context to leave.
   *
   * -1 to not reduce the number of lines.
   */
  size_t context;

  /**
   * \brief
   *   Storage for newly created diff blocks.
   *
   * If a diff block contains a big section of no-op lines, it needs
   * to be split into multiple blocks. When going through a diff wiith
   * diff_visit_blocks), we can't edit the list of diff blocks. Thus,
   * we cache the newly created blocks here for the caller to add to
   * the diff after we're done.
   *
   * You must initialize this with list_new() and clean it up.
   */
  list_t new_blocks;
};

/**
 * \brief
 *    Reduce the number of lines of context for a diff_block.
 *
 * Useful after running another filter which nullifies the effect of
 * certain lines in a block. If a single block contains a no-op
 * section with more than twice the number of lines of needed context,
 * the given block has to be split into multiple blocks.
 */
enum diff_visit_blocks_func_return mangle_condense_visit(diff_block_t block, struct mangle_condense *condense);

/**
 * \brief
 *   Reduce the number of lines of context for an diff.
 *
 * Wraps around mangle_condense_visit() to apply that function to the
 * whole diff and readd the newly generated blocks to the diff.
 *
 * Useful after running another filter which nullifies the effects of
 * certain lines in a block. If a single block contains a no-op
 * section with more than twice the number of lines of needed context,
 * the given block has to be split into multiple blocks.
 *
 * \param diff The diff to condense.
 * \param lines_context The number of lines of context to condense the
 *   diff into.
 */
int mangle_condense(diff_t diff, size_t lines_context);

/**
 * \brief
 *   A filter function for mangle_killchange_visit().
 *
 * \see mangle_killchange()
 *
 * \param
 *   A string to examine and produce a new, filtered string from.
 * \param
 *   A pointer to custom data.
 * \return
 *   A new string based on in which may be free()d.
 */
typedef char *(*mangle_killchange_filter_func_t)(const char *in, void *data);

/**
 * \brief
 *   The parameters that may be passed to mangle_killchange().
 */
struct mangle_killchange
{
  /**
   * \brief
   *   A function to filter each diff line through before checking if
   *   a DIFF_LINE_MINUS and DIFF_LINE_PLUS pair of lines is equal for
   *   the purposes of mangle_killchange_visit().
   */
  mangle_killchange_filter_func_t filter_func;
  /**
   * \brief
   *   A generic/custom pointer to pass to the filter_func as its data
   *   (second) argument.
   */
  void *filter_func_data;
};

/**
 * \brief
 *   Replaces lines which have not changed according to filter_func
 *   changes with NOOP lines.
 */
enum diff_visit_blocks_func_return mangle_killchange_visit(diff_block_t block, struct mangle_killchange *data);

/**
 * \brief
 *   Replace certain changes in a diff with DIFF_LINE_NOOP lines.
 *
 * This function goes through all of a diff's blocks removing
 * changes. A filtering function must be provided which will be run
 * each DIFF_LINE_MINUS and DIFF_LINE_PLUS line. This filter function
 * must come up with the same string for both lines if the two lines
 * to be combined into one DIFF_LINE_NOOP line. For example, if the
 * filter function produces a new string which has replaced all areas
 * of whitespace with a single space character, then an effect similar
 * to running diff -b is achieved.
 *
 * \param diff
 *   The diff to condense.
 * \param filter_func
 *   A filter to be used to determine if two lines are ``equal'' and
 *   should be NOOPed.
 * \param filter_func_data
 *   The pointer to pass to filter_func as the second data argument.
 */
int mangle_killchange(diff_t diff, mangle_killchange_filter_func_t filter_func, void *filter_func_data);

/**
 * \brief
 *   A filter for mangle_killchange() which ignores _all_ whitespace
 *   changes.
 *
 * Satisfies the mangle_killchange_filter_func_t interface.
 *
 * Removes the portions of a string that diff -w would ignore.
 *
 * \see mangle_killchange_visit()
 *
 * \param in
 *   The string from which to make a copy stripped of whitespace.
 * \param data
 *   Ignored.
 * \return
 *   A new string with the appropriate whitespace chars removed or
 *   NULL on error.
 */
char *mangle_killchange_whitespace_filter(const char *in, void *data);

/**
 * \brief
 *   A filter for mangle_killchange() which ignores changes in the
 *   amount of whitespace but still preserves the presence of
 *   whitespace.
 *
 * Satisfies the mangle_killchange_filter_func_t interface.
 *
 * \see mangle_killchange()
 *
 * \param in
 *   A string which I will make a copy of while ignoring changes in
 *   whitespace.
 * \param data
 *   Ignored.
 * \return
 *   A newly allocated string which may be freed with free() out of
 *   which changes in the amount of whitespace have been removed.
 */
char *mangle_killchange_whitespaceamount_filter(const char *in, void *data);

/**
 * \brief
 *   Replace expanded CVS keywords with unexpanded versions.
 *
 * A helper for mangle_killchange_visit().
 * \param in
 *   String to mangle. If NULL, will deallocate our internal state.
 * \param cvs_keyword_list
 *   A list of CVS keywords to eat. Each is a compiled regex_t.
 * \return
 *   A new string which must be free()d or NULL on error.
 */
char *mangle_killchange_cvs_keywords_filter(const char *in, list_t cvs_keyword_list);

#endif /* _DIFFFILTER_MANGLE_H */
