/*
 * Copyright (C) 2011 Nathan Phillip Brink <ohnobinki@ohnopublishing.net>
 *
 * This file is a part of difffilter.
 *
 * difffilter is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * difffilter is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License
 * along with difffilter.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef _DIFFFILTER_PATTERN_H
#define _DIFFFILTER_PATTERN_H

/**
 * @file Functions for initializing, freeing, and testing difffilter
 * patterns.
 *
 * Currently, the only functionality beyond TRE's matches support is
 * storing a flag that says this match is to be inverted.
 */

#include <tre/regex.h>

/**
 * A difffilter pattern.
 */
struct difffilter_pattern
{
  /**
   * The TRE 
   */
  regex_t preg;

  /**
   * 0 if this pattern marks files for exclusion.  1 if this pattern
   * marks files to be passed through (accepted).
   */
  short accept;
};
typedef struct difffilter_pattern *difffilter_pattern_t;

/**
 * \brief
 *   Allocate memory for and initialize a difffilter_pattern_t.
 *
 * \param pattern
 *   The POSIX Extended Regular Expression to pass on to TRE for interpretation.
 * \param accept
 *   Whether a file matching this pattern should be passed through or
 *   rejected/excluded.
 * \param extended
 *   Indicates that a pattern is an Extended Regular Expression rather
 *   than a Basic Regular Expression.
 */
difffilter_pattern_t difffilter_pattern_new(const char *pattern, short accept, short extended);

/**
 * Free and deinitialize a difffilter_pattern_t.
 *
 * Compatible signature with free() and list_dealloc_func_t.
 *
 * @param 
 */
void difffilter_pattern_free(difffilter_pattern_t difffilter_pattern);

/**
 * Prints out the result of a regerror() call.
 *
 * Utility function for difffilter_pattern_new() and others callling
 * reg* functions.
 *
 * @param preg caller must call regfree() on this.
 */
int diff_print_regerror(const char *pattern, reg_errcode_t errcode, const regex_t *preg);


#endif /* _DIFFFILTER_PATTERN_H */
